#include <types.h>
#include <ata.h>
#include <blkdev.h>
#include <major.h>

/*
 *     
 */
int reg_blkdev(u32 major,const char *name,
		int (*dev_req)(u32, u8, u32, u32, unsigned char *))
{
    blkdev[major].name = name;
    blkdev[major].dev_request = dev_req;
    return major;
}


int blkdev_io(u32 major, u32 minor, u8 cmd, u32 nblock, u32 count, u8 *buff)
{
    if(blkdev[major].dev_request(minor,cmd,nblock,count,buff) < 0) return -1;
    return 0;
}


/*
 *    
 */
int blkdev_init()
{

/*
 *   ATA
 */

	if(hd_init() != MAJOR_ATA) {
	    printf("init error\n");
	    return -1;
	}

	return 0;

}


/*          */
u32 block_to_lba(u32 start_block, int minor)
{
    u32 lba;
    u8 dev = GET_DEV(minor);
    u16 part = GET_PART(minor);

    if((start_block < 0) || (part > 4)) return -1;

    lba = start_block * (BLK_SIZE/BYTE_PER_SECT);
    if(part != 0) lba += DEV_PT(dev,(part-1)).sect_before;

    return lba;
}


int stat_blkdev(u32 major, u32 minor, u8 *buff)
{
    if(blkdev_io(major,minor,STAT,0,0,buff) < 0) return -1;
    return 0;
}


int read_blkdev(u32 major, u32 minor, u64 start, u64 count, u8 *buff)
{

    u32 start_lba,
	s_count,
	start_block,
	end_block,
	tail,
	num_block;

/*
 * start_block -     (0,1,2, ...)
 * start_lba -    
 * s_count -    
 * tail -      
 * num_block -    
 */


    device_info_t dev_i;
    u8 *cache_buff;

    if(stat_blkdev(major,minor,(u8 *)&dev_i) < 0) return -1;

    start_block = start/dev_i.block_size;
    end_block = (start+count)/dev_i.block_size;
    tail = start % dev_i.block_size;
    num_block = (end_block - start_block) + 1;

#ifdef DEBUG
    printf("  - %d \n",dev_i.block_size);
    printf("    () - %d\n",dev_i.blocks_num);
    printf("    - %d\n",start_block);
    printf("     ,  - %d\n",tail);
    printf("    - %d\n\n",num_block);
#endif

    cache_buff = (u8 *)malloc(num_block * dev_i.block_size);
    memset(cache_buff,0,num_block * dev_i.block_size);

    start_lba = block_to_lba(start_block,minor);
    if(start_lba < 0) return -1;
    s_count = num_block * (BLK_SIZE/BYTE_PER_SECT);

#ifdef DEBUG
    printf("      - %d\n",start_lba);
    printf("    - %d\n\n",s_count);
#endif

    if(blkdev_io(major,minor,READ,start_lba,s_count,cache_buff) < 0) return -1;

    memcpy(buff,cache_buff+tail,count);

    free(cache_buff);
    return 0;
}


int write_blkdev(int major, int minor, u32 start, u32 count, u8 *buff)
{
    u32 start_lba,
	s_count,
	start_block,
	end_block,
	tail,
	num_block;

/*
 * start_block -     (0,1,2, ...)
 * start_lba -    
 * s_count -    
 * tail -      
 * num_block -    
 */


    device_info_t dev_i;
    u8 *cache_buff;

    if(stat_blkdev(major,minor,(u8 *)&dev_i) < 0) return -1;

    start_block = start/dev_i.block_size;
    end_block = (start+count)/dev_i.block_size;
    tail = start % dev_i.block_size;
    num_block = (end_block - start_block) + 1;

    cache_buff = (u8 *)malloc(num_block * dev_i.block_size);
    memset(cache_buff,0,num_block * dev_i.block_size);

    start_lba = block_to_lba(start_block,minor);
    if(start_lba < 0) return -1;
    s_count = num_block * (BLK_SIZE/BYTE_PER_SECT);

    if(blkdev_io(major,minor,READ,start_lba,s_count,cache_buff) < 0) return -1;
    memcpy(cache_buff+tail,buff,count);
    if(blkdev_io(major,minor,WRITE,start_lba,s_count,cache_buff) < 0) return -1;

    free(cache_buff);

    return 0;
}
