/*
 *    ext2fs
 */

#ifdef DEBUG
    #include <stdio.h>
    #include <fcntl.h>
    #include <unistd.h>
#endif

#include <linux/ext2_fs.h>
#include <types.h>
#include <ata.h>
#include <blkdev.h>

/*      */
#define TEMP_SIZE_LIMIT 49152	// 12   4096 

    struct ext2_super_block sb;	//   !

    int BLKSIZE;	//    

    u8 buff_grp[4096];	//     
    u8 buff[4096];

    u32 major = 0;	/*   */
    u32 minor = 0;	/*   */
    u64 start = 0;	/*     */
    u64 count = 0;	/*    */

/*       ATA- */
void set_perm()
{
    ioperm(CH0,8,1);
    ioperm(CH1,8,1);
    ioperm(CH2,8,1);
    ioperm(CH3,8,1);

    return;
}

void release_perm()
{
    ioperm(CH0,8,0);
    ioperm(CH1,8,0);
    ioperm(CH2,8,0);
    ioperm(CH3,8,0);

    return;
}


/*   */
void read_sb()
{
    start = 1024;		/*  */
    count = sizeof(sb);		/*    */

    memset(&sb, 0, 1024);

    if(read_blkdev(major, minor, start, count, (u8 *)&sb) < 0) {
    	printf("Error read superblock\n");
	exit(-1);
    }

/*     (MAGIC-) */
    if(sb.s_magic != EXT2_SUPER_MAGIC) {
	perror("magic");
	exit(-1);
    }

#ifdef DEBUG
    printf("\nSuperblock info\n-----------\n");
    printf("Inodes count\t\t-\t%u\n",sb.s_inodes_count);
    printf("Blocks count\t\t-\t%u\n",sb.s_blocks_count);
    printf("Block size\t\t-\t%u\n",1024 << sb.s_log_block_size);
    printf("First inode\t\t-\t%d\n",sb.s_first_ino);
    printf("Magic\t\t\t-\t0x%X\n",sb.s_magic);
    printf("Inode size\t\t-\t%d\n",sb.s_inode_size);
    printf("Inodes per group\t-\t%u\n",sb.s_inodes_per_group);
    printf("Blosks per group\t-\t%u\n",sb.s_blocks_per_group);
    printf("First data block\t-\t%u\n\n",sb.s_first_data_block);
#endif

    return;
}


/*    */
void read_gd()
{

/*     */
    BLKSIZE = 1024 << sb.s_log_block_size;

/*   ,      */
    start = (sb.s_first_data_block + 1) * BLKSIZE;
    count = BLKSIZE;

/*    */
    if(read_blkdev(major, minor, start, count, buff_grp) < 0) {
    	printf("Error read group descriptor table\n");
	exit(-1);
    }

    return;
}


/*   inode   i_num */
void get_inode(int i_num, struct ext2_inode *in)
{

/*
 * i_num -  inode 
 */

    struct ext2_group_desc gd;
    u64 group,
	index; //index -    inode_table

/*  ,    inode   i_num */
    group = (i_num - 1) / sb.s_inodes_per_group;

    memset((void *)&gd, 0, sizeof(gd));
    memcpy((void *)&gd, buff_grp + group*(sizeof(gd)), sizeof(gd));

/*    inode */
    index = (i_num - 1) % sb.s_inodes_per_group;

    start = (u64)(gd.bg_inode_table) * BLKSIZE + index * sb.s_inode_size;
    count = sb.s_inode_size;

    if(read_blkdev(major, minor, start, count, (u8*)in) < 0) {
    	printf("Error read inode\n");
	exit(-1);
    }

    return;
}


/*    # (blknum)   inode */
void read_iblock(struct ext2_inode *i, int blknum)
{

    start = (u64)(i->i_block[blknum])*BLKSIZE;
    count = BLKSIZE;

    if(read_blkdev(major, minor, start, count, buff) < 0) {
    	printf("Error read info block\n");
	exit(-1);
    }

    return;
}


/*    */
void get_root_dentry()
{
    int outf;
    struct ext2_inode in;

    get_inode(EXT2_ROOT_INO, &in);
    read_iblock(&in, 0);

#ifdef DEBUG
    outf = open("root.dent",O_CREAT|O_RDWR,0600);
    write(outf,buff,BLKSIZE);
    close(outf);
#endif

    return;
}


/*   inode    */
int get_i_num(char *name)
{

#ifdef DEBUG
    int i = 0;
#endif

    int rec_len = 0, name_len = 0;
    struct ext2_dir_entry_2 dent;

    name_len = strlen(name);
    if((!name_len) || (name_len > EXT2_NAME_LEN)) return -1;

    for(;;) {
	memcpy((void *)&dent, (buff + rec_len), sizeof(dent));

#ifdef DEBUG
    printf("Record # %d\n",i++);
    printf("Record lenght - %d\n", dent.rec_len);
    printf("File name - %s\n", dent.name);
    printf("Name lenght- %d\n", dent.name_len);
#endif

	if(!dent.rec_len) return (-1);
	if(!memcmp(dent.name, name, strlen(name))) break;
	rec_len += dent.rec_len;
    }

    return dent.inode;
}



/*  ATA  */
int ata_init()
{
    int i = 0;
    u8 dev = GET_DEV(minor);	/*   (0,1,2,3) */

    if(blkdev_init() < 0) return -1;

    if(DEV_STAT(dev) != ATA) {
	perror("device type");
	exit(-1);
    }

#ifdef DEBUG
/*       0 */
    for(; i < 4; i++) {
	printf("\n  %d - 0x%x\n",i,DEV_PT(dev,i).type_part);
	printf("  - 0x%x\n",DEV_PT(dev,i).bootable);
	printf("   %d - %d\n",i,DEV_PT(dev,i).sect_total);
	printf("  %d   - %d\n",i,DEV_PT(dev,i).sect_total/(BLK_SIZE/512));
	printf("   %d - %d\n\n",i,DEV_PT(dev,i).sect_before);
    }
#endif

    return 0;
}


/*     */
int read_file_blocks(struct ext2_inode *in, u8 *data_buff, u32 *num, u32 seek)
{
    int i = 0, n = 0;
    u16 start_block, end_block, num_block, tail;
    u8 *cache_buff;

/*
 * start_block -  
 * end_block -  
 * num_block -    
 * tail -   
 */


    if(seek >= (in->i_size)) return -1;
    if(seek >= TEMP_SIZE_LIMIT) return -1;
    if(((*num) + seek) > (in->i_size)) (*num) = (in->i_size) - seek;
    if(((*num) + seek) > TEMP_SIZE_LIMIT) (*num) = TEMP_SIZE_LIMIT - seek;

    start_block = seek/BLKSIZE;
    end_block = (seek + (*num))/BLKSIZE;
    tail = seek % BLKSIZE;
    num_block = (end_block - start_block) + 1;

    cache_buff = (u8 *)malloc(num_block * BLKSIZE);
    memset(cache_buff, 0, num_block * BLKSIZE);

    i = start_block;

    for(; n < num_block; n++) {
	read_iblock(in,i);
	memcpy((cache_buff + (n * BLKSIZE)), buff, BLKSIZE);
	i += 1;
    }

    memcpy(data_buff, (cache_buff + tail), (*num));

    free(cache_buff);
    return 0;
}

    

int ext2_read_file(u32 maj_num, u32 min_num, u8 *full_path, u8 *data_buff, u32 *num, u32 seek)
{

/*
 *  :
 * full_path -    
 * data_buff -   
 * num -     
 * seek -   
 */
 
    struct ext2_inode in;
    unsigned char tmp_buff[EXT2_NAME_LEN];
    static int i = 1;
    int n, i_num, type;


    if(full_path[0] != '/') {
	perror("slash");
	exit(-1);
    }

    major = maj_num;
    minor = min_num;

    set_perm();

    if(ata_init() < 0) {
	perror("ata_init");
	exit(-1);
    }

/*
 *  ,   ,
 *     
 */
    read_sb();
    read_gd();
    get_root_dentry();

    while(1) {

        memset(tmp_buff, 0, sizeof(tmp_buff));

	for(n = 0 ; n < EXT2_NAME_LEN; n++, i++) {
	    tmp_buff[n] = full_path[i];
	    if((tmp_buff[n] == '/') || (tmp_buff[n] == '\0')) {
		i++;
		break;
	    }
	}
	tmp_buff[n] = '\0';

	i_num = get_i_num(tmp_buff);
	if(i_num < 0) {
	    printf("No such file\n");
	    exit(-1);
	}
	get_inode(i_num, &in);

#ifdef DEBUG
	printf("Inode number - %u\n", i_num);
	printf("File name - %s\n", tmp_buff);
	printf("File size - %u\n",in.i_size);
#endif

	type = ((in.i_mode & 0xF000) >> 12);

	if(type & 0x04) {
	    read_iblock(&in, 0);
	    continue;
	}

	if(type & 0x08) {
	    if(read_file_blocks(&in, data_buff, num, seek) < 0) return -1;
	    break;
	}
    }

    release_perm();
    return 0;
}
