#include <types.h>
#include <ata.h>
#include <blkdev.h>
#include <major.h>


u16 channels[4] = { CH0, CH1, CH2, CH3 };

#define ATA_STATUS(x)	(channels[x] + 7)
#define ATA_CURRENT(x)	(channels[x] + 6)
#define ATA_HCYL(x)	(channels[x] + 5)
#define ATA_LCYL(x)	(channels[x] + 4)
#define ATA_SECTOR(x)	(channels[x] + 3)
#define ATA_NSECTOR(x)	(channels[x] + 2)
#define ATA_ERROR(x)	(channels[x] + 1)
#define ATA_DATA(x)	(channels[x])


/*
 *   struct ata_request    .
 */
struct ata_request {
    u8 dev;		/*   () : 0,1,2,3 */
    u16 *buff;		/*    */
    u32 nlba;		/*     r/w */
    u32 nsect;		/*    r/w */
    u8 err;		/*    */
    u8 lock;		/*     */
    u8 complite;	/*    () */
} dev_r;

#define CURRENT dev_r.dev


/* 
 *     
 */
#define OUT_P_B(val,port)	\
    asm(			\
	"outb %%al, %%dx"	\
	::"a"(val),"d"(port)	\
    )

#define OUT_P_W(val,port)	\
    asm(			\
	"outw %%ax, %%dx"	\
	::"a"(val),"d"(port)	\
    )

#define IN_P_B(val,port)	\
    asm(			\
	"inb %%dx, %%al"	\
	:"=a"(val)		\
	:"d"(port)		\
    )

#define IN_P_W(val,port)	\
    asm(			\
	"inw %%dx, %%ax"	\
	:"=a"(val)		\
	:"d"(port)		\
    )


void (*handler)(void);


int hd_busy(u8 dev)
{
    int t = 0;
    unsigned char status;

    do {
	t++;
	IN_P_B(status,ATA_STATUS(dev));
	if(t & TIMEOUT) break;
    } while (status & BSY);
    return t;
}


int hd_ready(u8 dev)
{
    int t = 0;
    unsigned char status;

    do {
	t++;
	IN_P_B(status,ATA_STATUS(dev));
	if(t & TIMEOUT) break;
    } while (!(status & DRDY));

    return t;
}


int hd_data_request(u8 dev)
{
    unsigned char status;

    IN_P_B(status,ATA_STATUS(dev));
    if(status & DRQ) return 1;
    return 0;
}


int check_error(u8 dev)
{
    unsigned char a;

    IN_P_B(a,ATA_STATUS(dev));
    if (a & ERR) return 1;
    return 0;
}


/*
 * e   
 */
int get_ata_identity(u8 dev)
{
    int i = 0;

    if(hd_busy(dev) & TIMEOUT) return -1;

    if((dev == 0) || (dev == 2))
        OUT_P_B(0xE0,ATA_CURRENT(dev));

    if((dev == 1) || (dev == 3))
        OUT_P_B(0xF0,ATA_CURRENT(dev));

    if(hd_ready(dev) == TIMEOUT) return -1;
    OUT_P_B(0xEC,ATA_STATUS(dev));

/*
 *  ATAPI  0xEC () .
 *     ERR -   ATAPI
 */
    if(check_error(dev)) return -1;

    for(; i < TIMEOUT; i++) if(hd_data_request(dev)) break;
    if(i == TIMEOUT) return -1;


/*
 *     
 */
    asm(
	"	cld			\n\t"
	"1:	inw %%dx, %%ax		\n\t"
	"	stosw			\n\t"
	"	decw %%cx		\n\t"
	"	jnz 1b			\n\t"
	::"D"(&DEV_ID(dev)), "d"(ATA_DATA(dev)),"c"(0x100));

    return 0;
}



/*     ATA,    */
void get_ata_info()
{
    int dev = 0;

    for(; dev < 4 ; dev++) {

/* 
 *   .    -   
 *  
 */
	if(hd_busy(dev) & TIMEOUT) {
	    DEV_STAT(dev) = 0;
	    continue;
	}

        DEV_STAT(dev) = ATA;
        if(get_ata_identity(dev) < 0) DEV_STAT(dev) = 0;

    }
}


/*
 *    
 */
void show_ata_info()
{
    int i = 0;

    for(; i < 4; i++) {

	printf("ATA%d - ",i);
	if(!DEV_STAT(i)) printf("none\n");
	if(DEV_STAT(i) == ATA) {
	    printf("exists\n");
	    printf("\tType - ATA Disk drive\n");
	    printf("\tModel - %s\n",DEV_ID(i).model);
	    printf("\tLBA capacipty - %d\n",DEV_ID(i).lba_capacity);
	}
    }
}



/* -    / */
void send_command(u8 cmd)
{

    if(hd_busy(CURRENT) & TIMEOUT) {
#ifdef DEBUG
	printf("Device busy\n");
#endif
	exit(-1);
    }

/*   (/)*/
/*   */
    if((CURRENT == 0) || (CURRENT == 2))
        OUT_P_B(0xE0|((dev_r.nlba & 0x0F000000) >> 24),ATA_CURRENT(CURRENT));

/*   */
    if((CURRENT == 1) || (CURRENT == 3))
        OUT_P_B(0xF0|((dev_r.nlba & 0x0F000000) >> 24),ATA_CURRENT(CURRENT));    

    hd_ready(CURRENT);

    OUT_P_B(dev_r.nsect,ATA_NSECTOR(CURRENT));
    OUT_P_B((dev_r.nlba & 0x000000FF),ATA_SECTOR(CURRENT));
    OUT_P_B(((dev_r.nlba & 0x0000FF00) >> 8),ATA_LCYL(CURRENT));
    OUT_P_B(((dev_r.nlba & 0x00FF0000) >> 16),ATA_HCYL(CURRENT));
    OUT_P_B(cmd,ATA_STATUS(CURRENT));

    return;
}


/*    c  */
void intr_read()
{
    int i = 0;

    dev_r.complite = 0;

    hd_busy(CURRENT);
    if(check_error(CURRENT)) {
	dev_r.err = 1;
	return;
    }

    while(!(hd_data_request(CURRENT))) continue;

    for(;i < 0x100; i++) 
	IN_P_W(dev_r.buff[i],ATA_DATA(CURRENT));

    dev_r.complite = 1;
    return;
}



/*      */
void intr_write()
{
    int i = 0;

    dev_r.complite = 0;

    hd_busy(CURRENT);
    if(check_error(CURRENT)) {
	dev_r.err = 1;
	return;
    }

    while(!(hd_data_request(CURRENT))) continue;

    for(;i < 0x100; i++) 
	OUT_P_W(dev_r.buff[i],ATA_DATA(CURRENT));

    dev_r.complite = 1;
    return;
}



int do_command(u32 count, u8 c)
{
    for(;;) {
	send_command(c);
	handler();
	while(!(dev_r.complite)) continue;
	if(dev_r.err) return -1;
	count--;
	if(!count) break;
	dev_r.nlba++;
	dev_r.buff += 0x100;
    }

    return 0;
}



int stat_hd(u16 part)
{

    device_info_t dev_i;

    dev_i.block_size = BLK_SIZE;
    dev_i.blocks_num = DEV_ID(CURRENT).lba_capacity/(BLK_SIZE/BYTE_PER_SECT);

    if(part != 0)
	dev_i.blocks_num = DEV_PT(CURRENT,(part-1)).sect_total/(BLK_SIZE/BYTE_PER_SECT);

    memcpy(dev_r.buff, (u16 *)&dev_i, sizeof(device_info_t));

    dev_r.lock = 0;

    return 0;
}


/*
 * -
 */
int hd_request(u32 minor, u8 cmd, u32 start_sect, u32 s_count, u8 *buff)
{

    u8 comm = 0;
    u16 part = GET_PART(minor);
    CURRENT = GET_DEV(minor);

/*      (  ATAPI) */
//    if(!(DEV_STAT(CURRENT) & 1) || (DEV_STAT(CURRENT) == ATAPI)) return -1;
    if(DEV_STAT(CURRENT) != ATA) return -1;

    if(part > 4) return -1;

/*
 *  :
 *    minor -   
 *    unsigned int start_block -    
 *    u32 count -   
 *    buff - ,   
 */

/*
 * ,      
 */
    while(dev_r.lock) continue;
    dev_r.lock = 1;
    dev_r.nlba = start_sect;	/*   */
    dev_r.nsect = 1;		/*  c   */
    dev_r.buff = (unsigned short *)buff;

    switch(cmd) {

	case STAT:
	    return stat_hd(part);
	    break;

	case READ:
	    comm = READ_FROM_DISK;	// 0x20;
	    handler = &intr_read;
	    break;

	case WRITE:
	    comm = WRITE_TO_DISK;	// 0x30;
	    handler = &intr_write;
	    break;

	default:
	    printf("Unknown command\n");
	    dev_r.lock = 0;
	    return -1;
	    break;
    }

    do_command(s_count, comm);

    dev_r.complite = 0; /*   " " */
    dev_r.lock = 0; /*   */

    return 0;
}



/*
 *        
 */
void get_pt_info()
{
    u8 dev;
    u32 minor = 0;
    int i = 0;
    unsigned char buff[0x200];

/*   ATA        */
    for(; i < 4; i++) {
	dev = GET_DEV(minor);
//	if(!(DEV_STAT(dev) & 1) || (DEV_STAT(dev) == ATAPI)) continue;
	if(DEV_STAT(CURRENT) != ATA) continue;
	if(hd_request(minor,READ,0,1,buff) < 0) break;
	memcpy(dev_status[dev].pt,(struct pt_struct *)(buff+0x1BE),PT_SIZE*4);
	minor += 0x100;
    }
    return;
}


/*    */
int hd_init()
{
    int i = 0, major = 0;

    get_ata_info();
    show_ata_info();
    get_pt_info();
    major = reg_blkdev(MAJOR_ATA,"ATA",&hd_request);

    return major;
}
